import os
import hashlib
import logging

import yaml

from raptor import repo
from raptor.utils import run


HOME = (os.environ.get('RAPTOR_HOME') or os.path.expanduser('~/.raptor'))
PACKS_HOME = os.path.join(HOME, 'buildpacks')
CACHE_HOME = os.path.join(HOME, 'cache')
BUILD_HOME = os.path.join(HOME, 'build')
CONFIG_FILE = os.path.join(HOME, 'config.yaml')


log = logging.getLogger(__name__)


class BuildPack(repo.Repo):

    def detect(self, app):
        """
        Given an app, run detect script on it to determine whether it can be
        built with this pack.  Return True/False.
        """
        script = os.path.join(self.folder, 'bin', 'detect')
        cmd = '%s %s' % (script, app.folder)
        result = run(cmd)
        return result.status_code == 0

    def compile(self, app):
        log.info('Compiling %s with %s' % (app.basename, self.basename))
        script = os.path.join(self.folder, 'bin', 'compile')

        cache_folder = os.path.join(CACHE_HOME, get_unique_repo_folder(app.url))

        cmd = ' '.join([script, app.folder, cache_folder])
        log.info(cmd)
        result = run(cmd)
        result.raise_for_status()

    def release(self, app):
        script = os.path.join(self.folder, 'bin', 'release')
        result = run('%s %s' % (script, app.folder))
        assert result.status_code == 0, ("Failed release on %s with %s "
                                         "buildpack" % (app, self.basename))
        return yaml.safe_load(result.output)

    def update(self, rev=None):
        """
        Override Repo.update to provide default rev when none is provided.
        """

        rev = rev or self.fragment or {
            'git': 'HEAD',
            'hg': 'tip',
        }[self.vcs_type]

        return super(BuildPack, self).update(rev)


class App(repo.Repo):
    """
    A Repo that contains a buildpack-compatible project.
    """

    def __init__(self, folder, url=None, buildpack=None, buildpack_order=None, *args,
                 **kwargs):
        super(App, self).__init__(folder, url, *args, **kwargs)
        # If buildpack is None here, we'll try self.detect_buildpack later.
        self._buildpack = buildpack
        self.buildpack_order = (buildpack_order or
                                get_config().get('buildpack_order'))

    def detect_buildpack(self):
        """
        Loop over installed build packs and run each one's 'detect' script on
        the project until one succeeds, and set as self.buildpack.  Raise
        exception if nothing matches.
        """
        detected = next(
            (bp for bp in list_buildpacks(preferred_order=self.buildpack_order)
             if bp.detect(self)), None)
        if detected is None:
            raise ValueError("Cannot determine app's buildpack.")
        return detected

    @property
    def buildpack(self):
        if not self._buildpack:
            self._buildpack = self.detect_buildpack()
        return self._buildpack

    def compile(self):
        self.buildpack.compile(self)

    def release(self):
        return self.buildpack.release(self)


def list_buildpacks(packs_dir=PACKS_HOME, preferred_order=None):
    if not os.path.exists(packs_dir):
        run('mkdir -p %s' % packs_dir)

    preferred_order = preferred_order or None
    # if we have a preferred_order, then use that first, and filesystem order
    # second.
    buildpacks = os.listdir(packs_dir)
    if preferred_order:
        # This is inefficient.  It doesn't matter.
        new = []
        for bp in preferred_order:
            if bp in buildpacks:
                # something from the configured order is installed.  Append to
                # new list so it's in proper position, and delete from old
                # list.
                new.append(bp)
                buildpacks.remove(bp)
        # Anything left in 'buildpacks' must not have been specified in the
        # configured order, so just tack it on at the end
        new.extend(buildpacks)
        buildpacks = new

    return [BuildPack(os.path.join(packs_dir, d)) for d in buildpacks]


def add_buildpack(url, packs_dir=PACKS_HOME, vcs_type=None):
    # Check whether the pack exists
    dest = os.path.join(packs_dir, repo.basename(url))
    # If folder already exists, assume that we've already checked out the
    # buildpack there.
    # TODO: check for whether the buildpack in the folder is really the same as
    # the one we've been asked to add.
    if not os.path.exists(packs_dir):
        run('mkdir -p %s' % packs_dir)
    bp = BuildPack(dest, url, vcs_type=vcs_type)
    bp.update()
    return bp


def get_unique_repo_folder(repo_url):
    """
    Given a repository URL, return a folder name that's human-readable,
    filesystem-friendly, and guaranteed unique to that repo.
    """
    return '%s-%s' % (repo.basename(repo_url), hashlib.md5(repo_url).hexdigest())


# Some buildpacks assume that an app will always be built in the same folder.
# The Python buildpack does this by virtue of relying on virtualenv+pip which
# leave some symlinks (inside <cache_folder>/.heroku/venv/local/) that point to
# paths inside the build folder.  If we don't provide a consistent build folder
# to this buildpack, then we can't use the buildpack's caching feature.
def get_build_folder(app_url):
    return os.path.join(BUILD_HOME, get_unique_repo_folder(app_url))


class use_buildfolder(object):
    """Context manager for putting you into an app-specific build directory on
    enter and returning you to where you started on exit.
    """
    def __init__(self, app_url):
        self.app_url = app_url
        self.orig_path = os.getcwd()

    def __enter__(self):
        self.temp_path = get_build_folder(self.app_url)
        if not os.path.isdir(self.temp_path):
            os.makedirs(self.temp_path, 0770)
        os.chdir(self.temp_path)
        return self.temp_path

    def __exit__(self, type, value, traceback):
        os.chdir(self.orig_path)


def get_config(file_path=CONFIG_FILE):
    try:
        with open(file_path, 'rb') as f:
            return yaml.safe_load(f)
    except IOError:
        return {}



import redis

from raptor.utils import parse_redis_url

def test_parse_redis_url():
    r = redis.StrictRedis(**parse_redis_url('redis://localhost:6379/0'))
    expected ={'db': 0,
               'decode_responses': False,
               'encoding': 'utf-8',
               'encoding_errors': 'strict',
               'host': 'localhost',
               'password': None,
               'port': 6379,
               'socket_timeout': None}
    assert r.connection_pool.connection_kwargs == expected



import sys
import os
import subprocess
import shutil
import tempfile
import urlparse
import isodate
from datetime import datetime


class tmpdir(object):
    """Context manager for putting you into a temporary directory on enter
    and deleting the directory on exit
    """
    def __init__(self):
        self.orig_path = os.getcwd()

    def __enter__(self):
        self.temp_path = tempfile.mkdtemp()
        os.chdir(self.temp_path)
        return self.temp_path

    def __exit__(self, type, value, traceback):
        os.chdir(self.orig_path)
        shutil.rmtree(self.temp_path, ignore_errors=True)


class chdir(object):
    def __init__(self, folder):
        self.orig_path = os.getcwd()
        self.temp_path = folder

    def __enter__(self):
        os.chdir(self.temp_path)

    def __exit__(self, type, value, traceback):
        os.chdir(self.orig_path)


class CommandException(Exception):
    """
    Custom exception class for displaying nice input from failed commands.
    Accepts an CommandResult object on init.
    """

    def __init__(self, result):
        message = ("Command '%(command)s' failed with status code "
                   "%(status_code)s.\n"
                   "output: %(output)s\n") % {
                       'command': result.command,
                       'status_code': result.status_code,
                       'output': result.output,
                   }
        super(CommandException, self).__init__(message)


class CommandResult(object):
    def __init__(self, command, output, status_code):
        self.command = command
        self.output = output
        self.status_code = status_code

    def __repr__(self):
        return '<CommandResult: %s,%s>' % (self.status_code, self.command)

    def raise_for_status(self):
        if self.status_code != 0:
            raise CommandException(self)


def run(command, verbose=False):
    """
    Run a shell command.  Capture the stdout and stderr as a single stream.
    Capture the status code.

    If verbose=True (the default), then print command and the output to the
    terminal.

    """
    p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT)
    output = ""
    status_code = None
    if verbose:
        print "run: %s" % command
    while status_code is None:
        status_code = p.poll()
        line = p.stdout.readline()
        if verbose:
            sys.stdout.write(line)
        output += line
    return CommandResult(command, output, status_code)


def parse_redis_url(url):
    """
    Given a url like redis://localhost:6379/0, return a dict with host, port,
    and db members.
    """
    parsed = urlparse.urlsplit(url)
    return {
        'host': parsed.hostname,
        'port': parsed.port,
        'db': int(parsed.path.replace('/', '')),
    }


utc = isodate.FixedOffset(0, 0, 'UTC')

def utcfromtimestamp(ts):
    """
    Given a UNIX timestamp, return a Python datetime with the tzinfo explicitly
    set to UTC (as opposed to datetime.utcfromtimestamp, which returns a naive
    datetime with UTC values).
    """
    dt = datetime.utcfromtimestamp(ts)
    return dt.replace(tzinfo=utc)



import errno
import posixpath
import re

from deployment.balancer.base import SshBasedBalancer

BACKEND_TPL = '{.backend={.host="%(host)s"; .port="%(port)s";}}'

DIRECTOR_TPL = """
// This pool was written by Velociraptor.  You probably shouldn't edit it.
director %(pool)s round-robin {
%(backends)s
}
"""


def _render_backend(node):
    host, port = node.split(':')
    return BACKEND_TPL % {
        'host': host,
        'port': port,
    }


def pool_to_str(pool, nodes):
    # Return an empty string if there are no nodes.  A content-less include
    # file is probably OK, but directives with no contents probably aren't.
    if not nodes:
        return ''

    return DIRECTOR_TPL % {
        'backends': '  ' + '\n  '.join(_render_backend(n) for n in nodes),
        'pool': pool
    }


def str_to_pool(pool_str):
    # parse out the hosts and ports
    hs_and_ps = re.findall('host="(.*?)"; .port="(.*?)"', pool_str,
                           re.MULTILINE)
    # hs_and_ps will look like [('a', '50'), ('b', '9000')]
    nodes = ['%s:%s' % pair for pair in hs_and_ps]
    pool = re.search('director (.*?) ', pool_str).group(1)
    return pool, nodes

# I think we're gonna have to have a master raptor-pools.vcl include file that
# includes each pool.  The alternative is to try to parse the whole file and
# reconstruct it, which sounds risky.

# There's a lot of stuff in the nginx balancer that we're going to want in the
# varnish one, for connecting to SSH.  That'll need to go into a base class or
# helper functions.


class VarnishBalancer(SshBasedBalancer):

    def __init__(self, config):
        self.include_dir = config.get('include_dir',
                                      '/etc/varnish/')
        self.reload_cmd = config.get('reload_cmd',
                                     '/etc/init.d/varnish reload')
        self.main_pool_file = config.get('main_pool_file', 'pools.vcl')
        self.pool_file_prefix = config.get('pool_file_prefix', 'pool-')
        super(VarnishBalancer, self).__init__(config)

    def _get_poolfile(self, pool):
        return posixpath.join(self.include_dir, self.pool_file_prefix + pool +
                              '.vcl')

    def _get_host_nodes(self, host, pool):
        # Return set of nodes currently configured in a given host and pool

        filename = self._get_poolfile(pool)
        try:
            contents = self._read_file(host, filename)
            poolname, nodes = str_to_pool(contents)
        except IOError as e:
            # It's OK if the file doesn't exist.  But other IOErrors should be
            # raised normally.
            if e.errno == errno.ENOENT:
                nodes = []
            else:
                raise
        return set(nodes)

    def _get_include_line(self, pool):
        return 'include "%s";' % self._get_poolfile(pool)

    def _set_host_nodes(self, host, pool, nodes):
        # Write the pool file
        filename = self._get_poolfile(pool)
        contents = pool_to_str(pool, nodes)
        self._write_file(host, filename, contents)

        # Also ensure that there's a root include file that includes the one we
        # just wrote.
        main_file = posixpath.join(self.include_dir, self.main_pool_file)
        try:
            main_contents = self._read_file(host, main_file)
        except IOError as e:
            if e.errno == errno.ENOENT:
                main_contents = ''
            else:
                raise
        include_line = self._get_include_line(pool)
        if not include_line in main_contents:
            self._write_file(host, main_file, main_contents + '\n' +
                             include_line)



import logging
import posixpath
import random
import string
import warnings
import abc

import paramiko
from django.conf import settings


class Balancer(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self, config):
        pass

    @abc.abstractmethod
    def add_nodes(self, pool_name, nodes):
        pass

    @abc.abstractmethod
    def delete_nodes(self, pool_name, nodes):
        pass

    @abc.abstractmethod
    def get_nodes(self, pool_name):
        pass


class SshBasedBalancer(Balancer):
    """
    A helper class for writing balancer backends that are configured by SSHing
    to the balancer hosts, writing config files, and reloading config.

    Subclasses should implement _set_host_nodes and _get_host_nodes methods
    that handle reading and writing config files (but not reloading).
    """
    def __init__(self, config):
        self.user = config.get('user', settings.DEPLOY_USER)
        self.password = config.get('password', settings.DEPLOY_PASSWORD)
        self.hosts = config.get('hosts', ['localhost'])
        self.tmpdir = config.get('tmpdir', '/tmp')

    def _get_connection(self, host):
        con = paramiko.SSHClient()
        con.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        con.connect(host, username=self.user, password=self.password)
        return con

    def _read_file(self, host, path):
        # create connection
        con = self._get_connection(host)
        # read file with SFTP
        sftp = con.open_sftp()
        f = sftp.open(path, 'rb')
        contents = f.read()
        # close connection
        f.close()
        con.close()
        # return file contents
        return contents

    def _sudo(self, con, cmd):
        stdin, stdout, stderr = con.exec_command('sudo ' + cmd)
        stdin.write(self.password + '\n')
        stdin.flush()
        logging.info(stdout.read())

    def _write_file(self, host, path, contents):
        con = self._get_connection(host)
        # read file with SFTP
        sftp = con.open_sftp()
        # write file to temporary location
        tmppath = posixpath.join(self.tmpdir,
                                 ''.join(random.choice(string.lowercase) for x
                                         in xrange(10)))
        f = sftp.open(tmppath, 'wb')
        f.write(contents)
        f.close()
        sftp.chmod(tmppath, 0644)
        # run sudo cmd to copy file to production location, and set owner/perms
        self._sudo(con, 'mv %s %s' % (tmppath, path))
        self._sudo(con, 'chown root %s' % path)
        # close SSH connection
        con.close()

    def _reload_config(self, host):
        # connect
        con = self._get_connection(host)
        # run sudo cmd to reload config
        self._sudo(con, self.reload_cmd)
        # disconnect
        con.close()

    def _delete_host_nodes(self, host, pool, nodes):
        current = self._get_host_nodes(host, pool)
        correct = current.difference(nodes)
        self._set_host_nodes(host, pool, list(correct))

    def get_nodes(self, pool):
        # Look at the config files of all the hosts.  Raise a warning about any
        # mismatched pools.
        nodes = None
        for host in self.hosts:
            host_nodes = self._get_host_nodes(host, pool)
            if nodes is None:
                nodes = host_nodes
            elif host_nodes != nodes:
                warnings.warn('Host %s has nodes %s for pool %s, but other '
                              'hosts have %s' % (host, pool,
                                                 str(list(host_nodes)),
                                                 str(list(nodes))))
                nodes = nodes.union(host_nodes)
        if nodes:
            return list(nodes)
        return []

    def add_nodes(self, pool, nodes):
        nodeset = set(self.get_nodes(pool)).union(nodes)
        for host in self.hosts:
            self._set_host_nodes(host, pool, list(nodeset))
            self._reload_config(host)

    def delete_nodes(self, pool, nodes):
        for host in self.hosts:
            self._delete_host_nodes(host, pool, nodes)
            self._reload_config(host)



import os

import pytest

from raptor import repo, build
from raptor.utils import tmpdir, run
from raptor.tests import tmprepo


def test_hg_folder_detection():
    with tmpdir():
        folder = os.path.abspath('.hg')
        run('mkdir -p %s' % folder)

        assert repo.guess_folder_vcs(os.getcwd()) == 'hg'


def test_git_folder_detection():
    with tmpdir():
        folder = os.path.abspath('.git')
        run('mkdir -p %s' % folder)

        assert repo.guess_folder_vcs(os.getcwd()) == 'git'


def test_svn_folder_detection():
    with tmpdir():
        folder = os.path.abspath('.svn')
        run('mkdir -p %s' % folder)

        assert repo.guess_folder_vcs(os.getcwd()) == 'svn'


def test_git_scheme_detection():
    url = 'git://github.com/heroku/heroku-buildpack-python'
    assert repo.guess_url_vcs(url) == 'git'


def test_git_suffix_detection():
    url = 'https://github.com/heroku/heroku-buildpack-python.git'
    assert repo.guess_url_vcs(url) == 'git'


def test_basename():
    url = 'https://github.com/heroku/heroku-buildpack-python.git'
    assert repo.basename(url) == 'heroku-buildpack-python'

def test_basename_fragment():
    url = 'https://github.com/heroku/heroku-buildpack-python.git#123456'
    assert repo.basename(url) == 'heroku-buildpack-python'



# TODO: Run local git/hg servers so we don't have to call out over the network
# during tests.
def test_hg_clone():
    url = 'https://bitbucket.org/btubbs/vr_python_example'
    with tmpdir():
        hgrepo = repo.Repo('hgrepo', url, 'hg')
        hgrepo.clone()
        assert hgrepo.get_url() == url


def test_git_clone():
    url = 'https://github.com/btubbs/vr_python_example.git'
    with tmpdir():
        gitrepo = repo.Repo('gitrepo', url, 'git')
        gitrepo.clone()
        assert gitrepo.get_url() == url


def test_version_in_fragment():
    rev = '16c1dba07ee78d5dbee1f965d91d3d61942ccb67'
    url = 'https://github.com/btubbs/vr_python_example.git#' + rev
    with tmpdir():
        bp = build.BuildPack('bp', url, 'git')
        bp.clone()
        bp.update()
        assert bp.version == rev


def test_hg_update():
    newrev = '13b6ce1e234a'
    oldrev = '496e15fd973f'
    with tmprepo('hg_python_app.tar.gz', 'hg') as r:
        r.update(newrev)
        f = 'newfile'
        assert os.path.isfile(f)

        r.update(oldrev)
        assert not os.path.isfile(f)


def test_git_update():
    newrev = '6c79fb7d071a9054542114eea70f69d5361a61ff'
    oldrev = '16c1dba07ee78d5dbee1f965d91d3d61942ccb67'
    with tmprepo('git_python_app.tar.gz', 'git') as r:
        r.update(newrev)
        f = 'newfile'
        assert os.path.isfile(f)

        r.update(oldrev)
        assert not os.path.isfile(f)


def test_update_norev():
    with tmprepo('hg_python_app.tar.gz', 'hg') as r:
        # repo.update requires passing a revision
        with pytest.raises(TypeError):
            r.update()


def test_buildpack_update_norev():
    with tmprepo('buildpack_hello.tar.gz', 'git', build.BuildPack) as r:
        rev = 'd0b1df4838d51c694b6bba9b6c3779a5e2a17775'
        # Unlike the Repo superclass, buildpacks can call .update() without
        # passing in a revision, since we don't want to make users think about
        # buildpack versions if they don't have to.
        r.update()
        assert r.version == rev


def test_buildpack_update_rev():
    with tmprepo('buildpack_hello.tar.gz', 'git', build.BuildPack) as r:
        rev = '410a52780f6fd9d10d09d1da54088c03a0e2933f'
        # But passing in a rev needs to be supported still
        r.update(rev)
        assert r.version == rev


def test_hg_get_version():
    rev = '496e15fd973f'
    with tmprepo('hg_python_app.tar.gz', 'hg') as r:
        r.update(rev)
        assert r.version == rev


def test_git_get_version():
    rev = '16c1dba07ee78d5dbee1f965d91d3d61942ccb67'
    with tmprepo('git_python_app.tar.gz', 'git') as r:
        r.update(rev)
        assert r.version == rev


def test_basename_trailing_space():
    # Catches https://bitbucket.org/yougov/velociraptor/issue/10/
    url = "ssh://hg@bitbucket.org/yougov/velociraptor "
    assert repo.basename(url) == 'velociraptor'



import xmlrpclib
import json
from datetime import datetime

import isodate
import redis

from raptor.utils import utcfromtimestamp, parse_redis_url


class Host(object):
    """
    Abstraction over the per-host Supervisor xmlrpc interface.  This is used in
    the Velociraptor web procs, command line clients, and Supervisor event
    listeners.

    Should be initialized with a hostname and either an existing xmlrpc Server
    connection or a port number to where the RPC service is listening.

    If also initialized with a Redis connection or URL, proc info will be
    cached in Redis.

    Call host.get_procs() to get a list of Proc objects for all
    Supervisor-managed processes on the host.  Call it with check_cache=True to
    allow fetching proc info from the Redis cache.

    Call host.get_proc('name') to get a Proc object for the process named
    'name'.  Call it with check_cache=True to allow fetching proc info from the
    Redis cache.  If the host has no proc with that name, ProcError will be
    raised.
    """
    def __init__(self, name, rpc_or_port=9001, supervisor_username=None,
                 supervisor_password=None, redis_or_url=None,
                 redis_cache_prefix='host_procs', redis_cache_lifetime=600):
        self.name = name
        self.username = supervisor_username
        self.password = supervisor_password

        # Allow passing in an RPC connection, or a port number for making one
        if isinstance(rpc_or_port, int):
            if self.username:
                self.rpc = xmlrpclib.Server('http://%s:%s@%s:%s' %
                                            (self.username,self.password, name,
                                             rpc_or_port))
            else:
                self.rpc = xmlrpclib.Server('http://%s:%s' % (name, rpc_or_port))
        else:
            self.rpc = rpc_or_port
        self.supervisor = self.rpc.supervisor

        if isinstance(redis_or_url, redis.StrictRedis):
            self.redis = redis_or_url
        elif isinstance(redis_or_url, basestring):
            self.redis = redis.StrictRedis(**parse_redis_url(redis_or_url))
        else:
            self.redis = None
        self.cache_key = ':'.join([redis_cache_prefix, name])
        self.cache_lifetime = redis_cache_lifetime

    def get_proc(self, name, check_cache=False):
        if check_cache:
            # Note that if self.redis=None, and check_cache=True, an
            # AttributeError will be raised.
            cached_json = self.redis.hget(self.cache_key, name)
            if cached_json:
                return Proc(self, json.loads(cached_json))
            else:
                procs_dict = self._get_and_cache_procs()
        else:
            procs_dict = self._get_and_cache_procs()

        if name in procs_dict:
            return Proc(self, procs_dict[name])
        else:
            raise ProcError('host %s has no proc named %s' % (self.name, name))


    def _get_and_cache_procs(self):
        proc_list = self.supervisor.getAllProcessInfo()
        # getAllProcessInfo returns a list of dicts.  Reshape that into a dict
        # of dicts, keyed by proc name.
        proc_dict = {d['name']: d for d in proc_list}
        if self.redis:
            # Use pipeline to do hash clear, set, and expiration in same redis call
            with self.redis.pipeline() as pipe:

                # First clear all existing data in the hash
                pipe.delete(self.cache_key)
                # Now set all the hash values, dumped to json.
                dumped = {d: json.dumps(proc_dict[d]) for d in proc_dict}
                pipe.hmset(self.cache_key, dumped)
                pipe.expire(self.cache_key, self.cache_lifetime)
                pipe.execute()

        return proc_dict

    def get_procs(self, check_cache=False):
        if check_cache:
            unparsed = self.redis.hgetall(self.cache_key)
            if unparsed:
                all_data = {v: json.loads(v) for v in unparsed.values()}
            else:
                all_data = self._get_and_cache_procs()
        else:
            all_data = self._get_and_cache_procs()

        return [Proc(self, all_data[d]) for d in all_data]

    def shortname(self):
        return self.name.split(".")[0]

    def __repr__(self):
        info = {
            'cls': self.__class__.__name__,
            'name': self.name,
        }
        return "<%(cls)s %(name)s>" % info


class Proc(object):
    """
    A representation of a proc running on a host.  Must be initted with the
    hostname and a dict of data structured like the one you get back from
    Supervisor's XML RPC interface.
    """

    def __init__(self, host, data):
        self.host = host
        self._data = data

        # Be explicit, not magical, about which keys we expect from the data
        # and which attributes we set on the object.
        self.description = data['description']
        self.exitstatus = data['exitstatus']
        self.group = data['group']
        self.logfile = data['logfile']
        self.name = data['name']
        # When a timestamp field is inapplicable, Supevisor will put a 0 there
        # instead of a real unix timestamp.
        self.now = utcfromtimestamp(data['now']) if data['now'] else None
        self.pid = data['pid']
        self.spawnerr = data['spawnerr']
        self.start_time = utcfromtimestamp(data['start']) if data['start'] else None
        self.state = data['state']
        self.statename = data['statename']
        self.stderr_logfile = data['stderr_logfile']
        self.stdout_logfile = data['stdout_logfile']
        self.stop_time = utcfromtimestamp(data['stop']) if data['stop'] else None

        # The names returned from Supervisor have a bunch of metadata encoded
        # in them (at least until we can get a Supervisor RPC plugin to return
        # it).  Parse that out and set attributes.
        for k, v in self.parse_name(self.name).items():
            setattr(self, k, v)

        # We also set some convenience attributes for JS/CSS. It would be nice
        # to set those in the JS layer, but that takes some hacking on
        # Backbone.
        self.jsname = self.name.replace('.', 'dot')
        self.id = '%s-%s' % (self.host.name, self.name)

    @staticmethod
    def parse_name(name):
        try:
            app_name, version, recipe_name, rel_hash, proc_name, port = name.split('-')

            return {
                'app_name': app_name,
                'version': version,
                'recipe_name': recipe_name,
                'hash': rel_hash,
                'proc_name': proc_name,
                'port': int(port)
            }
        except ValueError:
            return {
                'app_name': name,
                'version': 'UNKNOWN',
                'recipe_name': 'UNKNOWN',
                'hash': 'UNKNOWN',
                'proc_name': name,
                'port': 0
            }

    @classmethod
    def name_to_shortname(cls, name):
        """
        In Celery tasks you often have a proc name, and want to send events
        including the proc's shortname, but you don't want to do a XML RPC call
        to get a full dict of data just for that.
        """
        return '%(app_name)s-%(version)s-%(proc_name)s' % Proc.parse_name(name)


    def __repr__(self):
        return "<Proc %s>" % self.name

    def shortname(self):
        return '%s-%s-%s' % (self.app_name, self.version, self.proc_name)

    def as_node(self):
        """
        Return host:port, as needed by the balancer interface.
        """
        return '%s:%s' % (self.host.name, self.port)

    def as_dict(self):
        data = {}
        for k, v in self.__dict__.items():
            if isinstance(v, (basestring, int)):
                data[k] = v
            elif isinstance(v, datetime):
                data[k] = v.isoformat()
            elif v is None:
                data[k] = v
        data['host'] = self.host.name
        return data

    def as_json(self):
        return json.dumps(self.as_dict())

    def start(self):
        try:
            self.host.supervisor.startProcess(self.name)
        except xmlrpclib.Fault as f:
            if f.faultString == 'ALREADY_STARTED':
                pass
            else:
                raise

    def stop(self):
        try:
            self.host.supervisor.stopProcess(self.name)
        except xmlrpclib.Fault as f:
            if f.faultString == 'NOT_RUNNING':
                pass
            else:
                raise

    def restart(self):
        self.stop()
        self.start()


class ProcError(Exception):
    """
    Raised when you request a proc that doesn't exist.
    """
    pass



# Tools for doing some simple (clone and pull) operations with repositories.
import os
import urlparse
import re
import logging

from raptor.utils import chdir, run


log = logging.getLogger(__name__)


def guess_url_vcs(url):
    """
    Given a url, try to guess what kind of VCS it's for.  Return None if we
    can't make a good guess.
    """
    parsed = urlparse.urlsplit(url)

    if parsed.scheme in ('git', 'svn'):
        return parsed.scheme
    elif url.endswith('.git'):
        return 'git'
    return None


def guess_folder_vcs(folder):
    """
    Given a path for a folder on the local filesystem, see what kind of vcs
    repo it is, if any.
    """
    try:
        contents = os.listdir(folder)
        vcs_folders = ['.git', '.hg', '.svn']
        found = next((x for x in vcs_folders if x in contents), None)
        # Chop off the dot if we got a string back
        return found[1:] if found else None
    except OSError:
        return None


class Repo(object):

    def __init__(self, folder, url=None, vcs_type=None):
        # strip trailing slash from folder if present
        if folder.endswith('/'):
            folder = folder[:-1]

        self.folder = folder

        vcs_type = vcs_type or guess_folder_vcs(folder) or guess_url_vcs(url)
        if vcs_type is None:
            raise ValueError('vcs type not guessable from folder (%s) or URL '
                             '(%s) ' % (folder, url))

        self.vcs_type = vcs_type

        if url is None and not os.path.isdir(folder):
            raise ValueError('Must provide repo url if folder does not exist')
        url = url or self.get_url()
        # Strip off fragment
        url, _, self.fragment = url.partition('#')

        # Strip off trailing slash
        if url.endswith('/'):
            url = url[:-1]
        self.url = url

    def run(self, command):
        r = run(command)
        r.raise_for_status()
        return r

    def get_url(self):
        """
        Assuming that the repo has been cloned locally, get its default
        upstream URL.
        """
        cmd = {
            'hg': 'hg paths default',
            'git': 'git config --get remote.origin.url',
        }[self.vcs_type]
        with chdir(self.folder):
            r = self.run(cmd)
        return r.output.replace('\n', '')

    def clone(self):
        log.info('Cloning %s to %s' % (self.url, self.folder))
        cmd = {
            'hg': 'hg clone %s %s' % (self.url, self.folder),
            'git': 'git clone %s %s' % (self.url, self.folder),
        }[self.vcs_type]
        self.run(cmd)

    def update(self, rev):
        # If folder doesn't exist, do a clone.  Else pull and update.
        if not os.path.exists(self.folder):
            self.clone()

        log.info('Updating %s from %s' % (self.folder, self.url))

        with chdir(self.folder):
            if self.vcs_type == 'hg':
                self.run('hg pull')
                self.run('hg up %s' % rev)
            elif self.vcs_type == 'git':
                self.run('git pull origin master')
                self.run('git checkout %s' % rev)

    @property
    def basename(self):
        return basename(self.url)

    @property
    def version(self):
        if self.vcs_type == 'hg':
            r = self.run('hg identify -i %s' % self.folder)
            return r.output.rstrip('+\n')
        elif self.vcs_type == 'git':
            with chdir(self.folder):
                r = self.run('git rev-parse HEAD')
            return r.output.rstrip()

    def __repr__(self):
        values = {'classname': self.__class__.__name__, 'folder':
                  os.path.basename(self.folder)}
        return "%(classname)s <%(folder)s>" % values


def basename(url):
    """
    Return the name of the folder that you'd get if you cloned 'url' into the
    current working directory.
    """
    # It's easy to accidentally have whitespace on the beginning or end of the
    # url.
    url = url.strip()

    url, sep, fragment = url.partition('#')
    # Remove trailing slash from url if present
    if url.endswith('/'):
        url = url[:-1]
    # Also strip .git from url if it ends in that.
    return re.sub('\.git$', '', url.split('/')[-1])



"""
Utilities for running commands and reading/writing files on remote hosts over
SSH.
"""

import os
import traceback
import tempfile
import shutil
import hashlib
import paramiko
import random
import string
import posixpath
import pkg_resources
import json
import re

from fabric.api import sudo, get, put, task, env
from fabric.contrib import files
from fabric import colors

from raptor.models import Proc

BUILDS_ROOT = '/apps/builds'
PROCS_ROOT = '/apps/procs'
RELEASES_ROOT = '/apps/releases'


@task
def upload_build(local_path, build_name, user='nobody'):

    remote_path = posixpath.join(BUILDS_ROOT, build_name)
    if not files.exists(remote_path):
        sudo('mkdir -p ' + remote_path)
    remote_procfile = posixpath.join(remote_path, 'Procfile')

    if files.exists(remote_procfile):
        colors.green('%s already on server.  No need to upload.' % remote_procfile)
        return


    # check if the build is needed.  upload if so.

    colors.green('Uploading build')
    # Make a random filename for the remote tmp file.
    tmp_path = posixpath.join('/tmp',
                              ''.join(random.choice(string.ascii_letters) for x
                                      in xrange(20)))
    put(local_path, tmp_path, use_sudo=True)

    colors.green('Unpacking build')

    tar_cmd = 'tar xjf {tmp_path} -C {remote_path} --strip-components 1'
    tar_cmd = tar_cmd.format(**vars())
    sudo(tar_cmd)
    sudo('chown -R {user} {remote_path}'.format(**vars()))
    sudo('chgrp -R admin {remote_path}'.format(**vars()))
    sudo('chmod -R g+w {remote_path}'.format(**vars()))
    # todo: mark directories as g+s

    colors.green('Deleting build tarball')
    sudo('rm ' + tmp_path)


@task
def upload_release(build_path, build_name, config_path, release_path, envsh_path, user='nobody'):
    # See if /apps/releases/<name> exists.  If not, make it.
    if not files.exists(release_path):
        colors.green('Creating remote directory')
        sudo('mkdir -p ' + release_path)

    upload_build(build_path, build_name)

    colors.green('Uploading config')
    # If we let Fabric upload settings.yaml, it might step on its own toes when
    # there are simultaneous deployments.  So handle that ourselves.
    with SSHConnection(env.host_string, env.user, env.password) as ssh:
        ssh.put_file(config_path, posixpath.join(release_path,
                                                 'settings.yaml'), owner=user)
        ssh.put_file(envsh_path, posixpath.join(release_path, 'env.sh'),
                     owner=user)


def parse_procfile(path):
    """
    Read a Procfile and return a dictionary with the proc names as keys and
    program locations as values.
    """
    with open(path, 'rb') as f:
        procs = dict([w.strip() for w in s.split(":", 1)] for s in f)
    return procs


def get_template(name):
    here = os.path.dirname(__file__)
    return os.path.join(here, 'templates', name)


class Deployer(object):
    mountpoints = []
    lxc_config_template = None

    def __init__(self, build_name, release_name, proc, port, user, use_syslog):
        self.build_name = build_name
        self.release_name = release_name
        self.proc = proc
        self.port = port
        self.user = user
        self.use_syslog = use_syslog
        self.proc_name = '-'.join([release_name, proc, str(port)])
        self.proc_path = posixpath.join(PROCS_ROOT, self.proc_name)
        self.proc_tmp_path = posixpath.join(self.proc_path, 'tmp')
        self.build_path = posixpath.join(BUILDS_ROOT, self.build_name)
        self.release_path = posixpath.join(RELEASES_ROOT, self.release_name)

        self.envsh_path = posixpath.join(self.proc_path, 'env.sh')
        self.yaml_settings_path = posixpath.join(self.proc_path,
                                                 "settings.yaml")
        self.proc_line = self.get_proc_line()

    def run(self):
        """
        Do everything needed to get the proc running on the remote host.
        """
        self.configure_proc()
        self.reload_supervisor()

    def configure_proc(self):
        """
        Wrap all the functions needed to get a proc configured on the remote
        host.
        """
        sudo('mkdir -p ' + self.proc_path)

        self.write_proc_conf()
        self.write_proc_lxc()
        self.write_start_proc_sh()
        self.write_start_proc_inner()
        self.create_proc_tmpdir()
        self.upload_uptester()
        self.create_mount_points()
        self.copy_config()

    def copy_config(self):
        """
        On the remote host, copy settings.yaml and env.sh from the release
        folder into the proc folder.
        """

        sudo('cp %(release_path)s/settings.yaml %(proc_path)s/' % self.__dict__)
        sudo('cp %(release_path)s/env.sh %(proc_path)s/' % self.__dict__)

    def get_paths(self):
        """
        Return a dictionary of paths to be used when writing the proc config.
        Must include keys for 'home', 'tmp', 'settings', and 'envsh'.
        """
        return {
            'home': self.proc_tmp_path,
            'tmp': self.proc_tmp_path,
            'settings': self.yaml_settings_path,
            'envsh': self.envsh_path,
        }

    def get_wrapper_cmd(self):
        """
        Return the command that should be run to actually start up the process.
        """
        return 'exec %s' % self.proc_line

    def reload_supervisor(self):
        # For this to work, the host must have PROCS_ROOT/*/proc.conf in
        # the files include line in the main supervisord.conf
        sudo('supervisorctl reread')
        sudo('supervisorctl add ' + self.proc_name)

    def create_mount_points(self):
        # Make container mount points.  lxc will handle the actual mounts,
        # but we need to make the folders.
        for m in self.mountpoints:
            sudo('mkdir -p %s%s' % (self.proc_path, m))

    def write_proc_conf(self):
        if self.use_syslog:
            stdout_log = "syslog"
        else:
            stdout_log = posixpath.join(self.proc_path, 'log')
        proc_conf_vars = {
            'release_name': self.release_name,
            'proc': self.proc,
            'port': self.port,
            'proc_path': self.proc_path,
            'stdout_log': stdout_log,
            'user': self.get_proc_conf_user(),
            'release_path': self.release_path,
        }
        proc_conf_tmpl = get_template('proc.conf')
        remote_supd = posixpath.join(self.proc_path, 'proc.conf')
        files.upload_template(proc_conf_tmpl, remote_supd, proc_conf_vars,
                              use_sudo=True)

    def write_proc_lxc(self):
        # write the lxc config.
        if self.lxc_config_template:
            lxc_tmpl = get_template(self.lxc_config_template)
            remote_lxc_path = posixpath.join(self.proc_path, 'proc.lxc')
            files.upload_template(lxc_tmpl, remote_lxc_path, {
                'build_path': self.build_path,
                'proc_path': self.proc_path,
            }, use_sudo=True)

    def write_start_proc_sh(self):
        """
        Write the outer script that sets up the container.
        """
        sh_script = get_template('start_proc.sh')
        sh_remote = posixpath.join(self.proc_path, 'start_proc.sh')
        tmpl_data = {'cmd': self.get_wrapper_cmd()}
        files.upload_template(sh_script, sh_remote, tmpl_data, use_sudo=True)
        sudo('chmod +x %s' % sh_remote)

    def write_start_proc_inner(self):
        """
        Write the inner script that runs inside the container, loads up env
        vars, and actually starts the proc.
        """
        sh_script = get_template('proc.sh')
        sh_remote = posixpath.join(self.proc_path, 'proc.sh')
        tmpl_data = self.get_paths()
        tmpl_data['cmd'] = self.proc_line
        tmpl_data['port'] = self.port
        files.upload_template(sh_script, sh_remote, tmpl_data, use_sudo=True)
        sudo('chmod +x %s' % sh_remote)

    def get_proc_conf_user(self):
        # Uncontained procs are just executed by Supervisor directly, so
        # proc.conf needs to be configured with the real username that we want
        # to run as
        return self.user

    def create_proc_tmpdir(self):
        # Create a place for the proc to stick temporary files if it needs to.
        sudo('mkdir -p ' + self.proc_tmp_path)
        sudo('chown %s %s' % (self.user, self.proc_tmp_path))

    def get_proc_line(self):
        try:
            # Make a temporary copy of the Procfile for parsing.  Fetch it from
            # the host so we don't even need to have a local copy of the build
            # in order to add a new deploy of an existing release.
            tempdir = tempfile.mkdtemp()
            local_procfile = posixpath.join(tempdir, 'Procfile')
            get(posixpath.join(self.build_path, 'Procfile'), local_procfile)

            procs = parse_procfile(local_procfile)
            if not self.proc in procs:
                proc_names = ', '.join(procs.keys())
                raise KeyError("%s not found in Procfile (%s defined)" %
                               (self.proc, proc_names))
            # From the parsed Procfile, pull out the command that should be
            # used to start the actual proc.
            return procs[self.proc]
        finally:
            # Clean up our tempdir
            shutil.rmtree(tempdir)

    def upload_uptester(self):
        """
        Upload libraptor's "uptester" program to the proc root.
        """
        uptester = pkg_resources.resource_filename('raptor',
                                                   'uptester/uptester')
        remote_path = posixpath.join(self.proc_path, 'uptester')
        put(uptester, remote_path, use_sudo=True)
        sudo('chmod +x %s' % remote_path)


class ContainedDeployer(Deployer):

    def get_proc_conf_user(self):
        # Contained procs will be started by Supervisor as root, and then su to
        # be the configured user once inside the container.
        return 'root'

    def get_paths(self):
        """
        Return a dictionary of paths to be used when writing the proc config.
        Must include keys for 'home', 'tmp', 'settings', and 'envsh'.
        """
        # In a container, these paths are always the same.
        return {
            'home': '/app',
            'tmp': '/tmp',
            'settings': '/settings.yaml',
            'envsh': '/env.sh',
        }

    def get_wrapper_cmd(self):
        container_name = self.proc_name
        lxc_config_path = posixpath.join(self.proc_path, 'proc.lxc')
        return build_container_cmd('/proc.sh', self.user,
                                   container_name, lxc_config_path)


class PreciseDeployer(ContainedDeployer):
    mountpoints = ('/app',
                   '/bin',
                   '/dev',
                   '/etc',
                   '/lib',
                   '/lib64',
                   '/opt',
                   '/usr',
                   '/proc',
                   '/run',
                   '/sys',
                   '/dev/pts',)

    lxc_config_template = 'precise.lxc'


@task
def configure_proc(build_name, release_name, proc, port, user='nobody', use_syslog=False,
                   contain=False):
    if not contain:
        deployer_cls = Deployer
    else:
        issue = sudo('cat /etc/issue')
        if '12.04' in issue:
            deployer_cls = PreciseDeployer
        else:
            raise ValueError("Could not determine deployer type for %s" %
                             issue)
    deployer_cls(build_name, release_name, proc, port, user, use_syslog).run()


@task
def deploy_parcel(build_path, config_path, envsh_path, swarm,
                  proc, port, user='nobody', use_syslog=False, contain=False,
                  release_hash=None):
    # Builds have timstamps, but releases really don't care about them.  Two
    # releases created at different times with the same build and settings
    # should be treated the same.  So throw away the timestamp portion of the
    # build name when creating the release name.
    build_file = os.path.basename(build_path)
    nameparts = build_file.split('-')
    app_name = nameparts[0]
    version = nameparts[1]

    build_name = '-'.join([app_name, version])

    # A release name should look like gryphon-2.2.3-ldc-d5338b8a07, where the
    # random looking chars at the end come from a hash of the build tarball and
    # settings.yaml used for the release.
    if release_hash is None:
        chars = open(build_path, 'rb').read() + open(config_path, 'rb').read()
        release_hash = hashlib.md5(chars).hexdigest()[:8]
    release_name = '-'.join([app_name, version, swarm, release_hash])
    release_path = posixpath.join(RELEASES_ROOT, release_name)

    # Ensure that the proc path is written before the release is uploaded, to
    # prevent a race condition with the scooper.
    proc_name = '-'.join([release_name, proc, str(port)])
    proc_path = posixpath.join(PROCS_ROOT, proc_name)
    sudo('mkdir -p ' + proc_path)

    upload_release(build_path, build_name, config_path, release_path,
                   envsh_path, user=user)

    configure_proc(build_name, release_name, proc, port, user,
                   use_syslog=use_syslog, contain=contain)


def build_container_cmd(cmd, user, container_name, lxc_config_path):
    """
    Build up the shell command needed to launch a program inside a container.
    This is used both in deployments and launching the uptester.
    """
    tmpl = """exec lxc-start --name %(container_name)s -f %(lxc_config_path)s -- su --preserve-environment --shell /bin/bash -c "cd /app;source /env.sh; exec %(cmd)s" %(user)s"""
    return tmpl % vars()


def build_contained_uptests_command(proc_path, proc, host, port, user):
    """
    Build the command string for uptesting the given proc inside its lxc
    container.
    """
    cmd = "/uptester %(folder)s %(host)s %(port)s" % {
        'folder': posixpath.join('/app/uptests', proc),
        'host': host,
        'port': port,
    }
    container_name = posixpath.basename(proc_path) + '-uptest'
    lxc_config_path = posixpath.join(proc_path, 'proc.lxc')
    return build_container_cmd(cmd, user, container_name, lxc_config_path)


def build_uncontained_uptests_command(build_path, proc_path, proc, host,
                                      port, user):
    """
    Build the command string for uptesting a proc that's not inside an lxc
    container.
    """
    tmpl = """su -c "cd %(build_path)s;source %(envsh_path)s;%(uptester)s %(folder)s %(host)s %(port)s" %(user)s"""
    return tmpl % {
        'release_path': build_path,
        'folder': posixpath.join(build_path, 'uptests', proc),
        'host': host,
        'port': port,
        'user': user,
        'uptester': posixpath.join(proc_path, 'uptester'),
    }


def build_uptests_command(build_path, proc_path, proc, host, port, user):
    """
    Pick from the available uptest command builders based on what's present in
    the proc and build dirs.  Run that builder and return its result.
    """
    lxc_conf_path = posixpath.join(proc_path, 'proc.lxc')
    if files.exists(lxc_conf_path):
        return build_contained_uptests_command(proc_path, proc,
                                               env.host_string, port, user)
    else:
        return build_uncontained_uptests_command(build_path, proc_path, proc,
                                                 env.host_string, port, user)


@task
def ensure_uptester(proc_path):
    # If there's no uptester in the proc folder, put one there.
    uptester_path = posixpath.join(proc_path, 'uptester')
    uptester = pkg_resources.resource_filename('raptor', 'uptester/uptester')
    if files.exists(proc_path) and not files.exists(uptester_path):
        put(uptester, uptester_path, use_sudo=True)
        sudo('chmod +x %s' % uptester_path)


@task
def run_uptests(proc, user='nobody'):
    procdata = Proc.parse_name(proc)
    procname = procdata['proc_name']
    build_name = '%(app_name)s-%(version)s' % procdata
    build_path = posixpath.join(BUILDS_ROOT, build_name)
    proc_path = posixpath.join(PROCS_ROOT, proc)

    tests_path = posixpath.join(build_path, 'uptests', procname)
    try:
        ensure_uptester(proc_path)
        if files.exists(tests_path):
            # determine whether we're running contained or uncontained
            cmd = build_uptests_command(build_path, proc_path, procname,
                                        env.host_string, procdata['port'],
                                        user)

            result = sudo(cmd)
            # Though the uptester emits JSON to stdout, it's possible for the
            # container or env var setup to emit some other output before the
            # uptester even runs.  Stuff like this:

            # 'bash: /app/env.sh: No such file or directory'

            # Split that off and prepend it as an extra first uptest result.
            # Since results should be a JSON list, look for any characters
            # preceding the first square bracket.

            m = re.match('(?P<prefix>[^\[]*)(?P<json>.*)', result, re.S)

            # If the regular expression doesn't even match, return the raw
            # string.
            if m is None:
                return [{
                    'Passed': False,
                    'Name': 'uptester',
                    'Output': result,
                }]

            parts = m.groupdict()
            try:
                parsed = json.loads(parts['json'])
                if len(parts['prefix']):
                    parsed.insert(0, {
                        'Passed': False,
                        'Name': 'uptester',
                        'Output': parts['prefix']
                    })
                return parsed
            except ValueError:
                # If we still fail parsing the json, return a dict of our own
                # with all the output inside.
                return [{
                    'Passed': False,
                    'Name': 'uptester',
                    'Output': result
                }]
        else:
            return []

    except (Exception, SystemExit):
        # Fabric will raise SystemExit if we don't supply the right password
        # and abort_on_prompts is True.  Here we catch any exception raised
        # during the uptests and pass it back in the same format as other test
        # results.
        return [{
            'Name': None,
            'Output': traceback.format_exc(),
            'Passed': False,
        }]


@task
def delete_proc(proc):
    if not proc:
        raise SystemExit("You must supply a proc name")
    # stop the proc
    sudo('supervisorctl stop %s' % proc)
    # remove the proc
    sudo('supervisorctl remove %s' % proc)

    # delete the proc dir
    proc_dir = posixpath.join(PROCS_ROOT, proc)
    if files.exists(proc_dir):
        sudo('rm -rf %s' % proc_dir)


@task
def delete_release(release, releases_root=RELEASES_ROOT, procs_root=PROCS_ROOT,
                   cascade=False):
    procs = get_procs()

    releases_in_use = set(['%(app_name)s-%(version)s-%(recipe_name)s-%(hash)s' %
                           Proc.parse_name(p) for p in procs])

    # see if there are procs pointing to this release
    # if so, and cascade==True, then delete procs pointing to this release.
    # Else fail.
    if release in releases_in_use:
        if cascade:
            # delete all procs for this release
            procs_to_delete = [p for p in procs if p.startswith(release)]
            for proc in procs_to_delete:
                delete_proc(proc)
        else:
            raise SystemExit('NOT DELETING %s. Release is currently in use, '
                             'and cascade=False' % release)
    sudo('rm -rf %s/%s' % (releases_root, release))


@task
def delete_build(build, builds_root=BUILDS_ROOT, releases_root=RELEASES_ROOT,
                 procs_root=PROCS_ROOT, cascade=False):
    build_releases = get_build_releases(build)
    if len(build_releases):
        if not cascade:
            raise SystemExit("NOT DELETING %s. Build is currently in use, "
                             "and cascade=False" % build)
        else:
            for release in build_releases:
                delete_release(release, releases_root, procs_root, cascade)
    sudo('rm -rf %s/%s' % (builds_root, build))


@task
def clean_releases_folders(releases_root=RELEASES_ROOT, procs_root=PROCS_ROOT):
    """ Check in releases_root for releases not being used by procs, so we can
    clean them up.

    You may choose not to execute the actual delete (to test for example), and
    if you choose to be verbose it will print out the releases it will delete.
    """

    if files.exists(procs_root, use_sudo=True) and \
        files.exists(releases_root, use_sudo=True):
        procs = get_procs()
        releases = get_releases()
        releases_in_use = set([
            '%(app_name)s-%(version)s-%(recipe_name)s-%(hash)s' %
            Proc.parse_name(p) for p in procs])
        deleted = []
        for release in releases:
            if release not in releases_in_use:
                deleted.append(release)
                delete_release(release, releases_root, procs_root, False)
        colors.green("Cleaned up %i releases." % len(deleted))


def clean_builds_folders(builds_root=BUILDS_ROOT, releases_root=RELEASES_ROOT):
    """
    Check in builds_root for builds not being used by releases.
    """

    if files.exists(builds_root, use_sudo=True) and files.exists(releases_root,
                                                                 use_sudo=True):
        releases = get_releases()
        builds = set(get_builds())

        builds_in_use = set([release_to_build(r) for r in releases])
        unused_builds = builds.difference(builds_in_use)
        for build in unused_builds:
            delete_build(build, builds_root, releases_root)


@task
def get_procs(procs_root=PROCS_ROOT):
    """
    Return the names of all the procs on the host.
    """
    return sudo('ls -1 %s' % procs_root).split()


@task
def get_releases(releases_root=RELEASES_ROOT):
    """
    Return the names of all the releases on the host.
    """
    return sudo('ls -1 %s' % releases_root).split()


@task
def get_builds(builds_root=BUILDS_ROOT):
    """
    Return the names of all the builds on the host.
    """
    return sudo('ls -1 %s' % builds_root).split()


@task
def get_release_procs(release, release_root=RELEASES_ROOT, procs_root=PROCS_ROOT):
    """
    Given a release name, return a list of all the procs on the host that use
    that release.
    """
    procs = get_procs()
    def proc_to_release(proc):
        """
        Given a proc name, return the name of the release that it uses.
        """
        tmpl = '%(app_name)s-%(version)s-%(recipe_name)s-%(hash)s'
        return tmpl % Proc.parse_name(p)
    return [p for p in procs if proc_to_release(p)==release]


def release_to_build(release):
    """
    Given a release name, return the name of the build that it uses.
    """
    try:
        return "%s-%s" % tuple(release.split('-')[:2])
    except ValueError:
        return "unknown-unknown"

@task
def get_build_releases(build, builds_root=BUILDS_ROOT,
                       releases_root=RELEASES_ROOT):
    releases = get_releases()
    return [r for r in releases if release_to_build(r)==build]


@task
def clean_releases_and_builds():
    """
    Delete any releases not in use by procs on the host.

    Then delete any builds not in use by releases on the host.
    """
    clean_releases_folders(RELEASES_ROOT, PROCS_ROOT)

    clean_builds_folders(BUILDS_ROOT, RELEASES_ROOT)


class SSHConnection(object):
    """
    A context manager for creating a non-Fabric SSH connection with a username and
    password, then ensuring it gets cleaned up.

    You can also instantiate this object without using it as a context manager.
    If you do, it's up to you to call close() when you're done.
    """
    # This class began life with grand ambitions to be a Fabric replacement.
    # Then I realized how much ugliness Fabric was hiding from me.  Now this
    # class is just used for writing the settings.yaml file inside a release,
    # because Fabric jobs would delete each other's temp files if two workers
    # happened to try deploying the same file at the same time
    def __init__(self, host, user, password, tmpdir='/tmp'):
        self.host = host
        self.user = user
        self.password = password
        self.tmpdir = tmpdir
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(self.host, username=self.user, password=self.password)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.ssh.close()

    def sudo(self, cmd):
        stdin, stdout, stderr = self.ssh.exec_command('sudo ' + cmd)
        stdin.write(self.password + '\n')
        stdin.flush()
        return stdout.read(), stderr.read()

    def write_file(self, path, contents, mode=0644, owner=None,
                   group=None):
        sftp = self.ssh.open_sftp()

        # There's no way to sftp.write as sudo, so we write to a temporary
        # location and then mv the file into place.
        tmppath = posixpath.join(self.tmpdir,
                                 ''.join(random.choice(string.ascii_letters) for x
                                         in xrange(10)))
        f = sftp.open(tmppath, 'wb')
        f.write(contents)
        f.close()
        sftp.chmod(tmppath, mode)
        # run sudo cmd to copy file to production location, and set owner/perms
        self.sudo('mv %s %s' % (tmppath, path))

        if owner:
            self.sudo('chown %s %s' % (owner, path))

        if group:
            self.sudo('chgrp %s %s' % (group, path))

    def put_file(self, local_path, remote_path, *args, **kwargs):
        return self.write_file(remote_path, open(local_path, 'rb').read(),
                               *args, **kwargs)



import os
import tempfile
import shutil
import xmlrpclib

from raptor import repo
from raptor.utils import run

# A fake Supervisor xmlrpc interface.
class FakeSupervisor(object):
    # Set this at test time in order to fake an exception.
    exception = None

    # Set this dict in order to control what data comes back from the
    # getAllProcessInfo and getProcessInfo calls.  (Slightly tweaked to include
    # a 'host' key.)
    # XXX A lot of tests rely on this bit of data.
    process_info = {
        'dummyproc': {
            'description': 'pid 5556, uptime 16:05:53',
            'exitstatus': 0,
            'group':  'dummyproc',
            'logfile':  '/var/log/supervisor/dummyproc-stdout---supervisor-cYv5Q2.log',
            'name':  'dummyproc',
            'now': 1355897986,
            'pid': 5556,
            'spawnerr':  '',
            'start': 1355897986,
            'state': 20,
            'statename':  'RUNNING',
            'stderr_logfile':  '/tmp/pub.log',
            'stdout_logfile':  '/var/log/supervisor/dummyproc-stdout---supervisor-cYv5Q2.log',
            'stop': 1355897986,
            'host': 'somewhere'},
        'node_example-v2-local-f96054b7-web-5003': {
            'description': 'Exited too quickly (process log may have details)',
            'exitstatus': 0,
            'group': 'node_example-v2-local-f96054b7-web-5003',
            'logfile': '/apps/procs/node_example-v2-local-f96054b7-web-5003/log',
            'name': 'node_example-v2-local-f96054b7-web-5003',
            'now': 1355955939,
            'pid': 0,
            'spawnerr': 'Exited too quickly (process log may have details)',
            'start': 1355898065,
            'state': 200,
            'statename': 'FATAL',
            'stderr_logfile': '/var/log/supervisor/node_example-v2-local-f96054b7-web-5003-stderr---supervisor-gL_lvl.log',
            'stdout_logfile': '/apps/procs/node_example-v2-local-f96054b7-web-5003/log',
            'stop': 1355898065}}

    def _fake_fault(self):
        if self.exception:
            raise self.exception

    def getProcessInfo(self, name):
        self._fake_fault()
        try:
            return self.process_info[name]
        except KeyError:
            # Raise exception like supervisor would when asked to return data
            # on nonexistent Proc.
            raise xmlrpclib.Fault(10, 'BAD_NAME: %s' % name)

    def getAllProcessInfo(self):
        self._fake_fault()
        return self.process_info.values()

class FakeRPC(object):
    def __init__(self):
        self.supervisor = FakeSupervisor()


class tmprepo(object):
    """
    Context manager for creating a tmp dir, unpacking a specified repo tarball
    inside it, cd-ing in there, letting you run stuff, and then cleaning up and
    cd-ing back where you were when it's done.
    """
    def __init__(self, tarball, vcs_type, repo_class=None):
        # Repo tarballs must be in the same directory as this file.
        here = os.path.dirname(os.path.abspath(__file__))
        self.tarball = os.path.join(here, tarball)
        self.vcs_type = vcs_type
        self.orig_path = os.getcwd()
        self.repo_class = repo_class or repo.Repo

    def __enter__(self):
        self.temp_path = tempfile.mkdtemp()
        os.chdir(self.temp_path)
        cmd = 'tar -zxf %s --strip-components 1' % self.tarball
        result = run(cmd)
        result.raise_for_status()
        return self.repo_class('./', vcs_type=self.vcs_type)

    def __exit__(self, type, value, traceback):
        os.chdir(self.orig_path)
        shutil.rmtree(self.temp_path, ignore_errors=True)



"""
This module provides a Supervisor 'event handler' conforming to the interface
at http://supervisord.org/events.html.  To use it, add a section like this to
your Supervisor configuration:

[eventlistener:proc_publisher]
command=proc_publisher
events=PROCESS_STATE,PROCESS_GROUP,TICK_60
environment=REDIS_URL="redis://localhost:6379/0"

You should *not* configure a stdout_logfile for this program (it messes up the
communication with Supervisor).

For each PROCESS_STATE and PROCESS_GROUP message received, the publisher will
put a json-encoded dict on the Redis pubsub specified in the eventlistener's
environment variables.  Velociraptor web procs will listen on this pubsub for
proc state changes and update dashboards in realtime.

The listener will also maintain a cache of the host's proc data in a Redis
hash.  This cache will be updated whenever an event is received from
Supervisor.  By configuring the eventlistener to receive TICK_60 events, the
cache will be updated once per minute, even if no process state changes have
occurred.
"""

import os
import sys
import json
import datetime
import socket

import redis
from supervisor import childutils

from raptor.utils import parse_redis_url
from raptor.models import Host, Proc


def main():
    if not 'SUPERVISOR_SERVER_URL' in os.environ:
        raise SystemExit('supervisor_events_publisher must be run as a '
                         'supervisor event listener')

    # Use Supervisor's own RPC interface to get full proc data on each process
    # state change, since the emitted events don't have everything we want.
    rpc = childutils.getRPCInterface(os.environ)

    rcon = redis.StrictRedis(**parse_redis_url(os.environ['REDIS_URL']))
    hostname = os.getenv('HOSTNAME', socket.getfqdn())
    log('proc_publisher starting with hostname %s' % hostname)
    host = Host(hostname, rpc_or_port=rpc, redis_or_url=rcon,)

    for e in EventStream(hostname):
        handle_event(e, host, 'proc_events')
        log(e.emit())


def handle_event(event, host, pubsub_channel):
    data = event.emit()
    if event.eventname.startswith('PROCESS_STATE'):
        proc_name = data['payload_headers']['processname']
        proc_data = host.get_proc(proc_name, check_cache=False).as_dict()

        # Include event type with all messages so listeners can ignore messages
        # they don't care about.
        proc_data.update(event=event.eventname)
        host.redis.publish(pubsub_channel, json.dumps(proc_data))

    elif event.eventname == 'PROCESS_GROUP_REMOVED':
        # Supervisor's events interface likes to call things 'process' or
        # 'group' or 'processname', but the xmlrpc interface gives
        # processes a 'name' and 'group'.  Velociraptor likes to use
        # 'name'.
        data['name'] = data['payload_headers']['groupname']
        # Parse out the bits of the name
        data.update(Proc.parse_name(data['name']))
        # also add in the 'id' field that procs have, since the dashboard will
        # rely on that to pluck out the JS proc model
        data['id'] = '%s-%s' % (host.name, data['name'])
        host.redis.publish(pubsub_channel, json.dumps(data))

    # No matter what kind of event we got, update the host's whole cache.
    host.get_procs(check_cache=False)


class Event(object):

    def __init__(self, headers, payload, hostname):
        # Save the raw data
        self.headers = headers
        self.payload = payload

        self.host = hostname

        # Parse out some useful bits
        self.eventname = headers['eventname']
        self.payload_headers, self.payload_data = childutils.eventdata(payload
                                                                       + '\n')
        if 'when' in self.payload_headers:
            utime = float(self.payload_headers['when'])
            self.time = datetime.datetime.utcfromtimestamp(utime)
        else:
            self.time = datetime.datetime.utcnow()

    def emit(self):
        data = {
            'event': self.eventname,
            'host': self.host,
            'payload_headers': self.payload_headers,
            'payload_data': self.payload_data,
            'time': self.time.isoformat(),
        }
        return data

    def __repr__(self):
        return '<Event %s>' % self.eventname


class EventStream(object):
    """
    Iterator over Supervisor-emitted events.

        es = EventStream()
        for e in es:
            do_something(e)
    """

    def __init__(self, hostname):
        self.hostname = hostname
        # Makes testing easier
        self.stdin = sys.stdin
        self.stdout = sys.stdout
        self.stderr = sys.stderr

        # Flag for whether an 'ok' is needed at the beginning of the next
        # self.next() call.
        self._needs_ok = False


    def __iter__(self):
        while 1:
            yield self.next()

    def next(self):
        if self._needs_ok:
            childutils.listener.ok(self.stdout)

        headers, payload = childutils.listener.wait(self.stdin, self.stdout)
        self._needs_ok = True
        return Event(headers, payload, self.hostname)


def log(*args):
    """
    Since the Supervisor event listener spec attaches special meaning to
    stdout, we can't use normal print.  Use this instead to print to
    stderr.
    """
    for s in args:
        sys.stderr.write(str(s) + '\n')
    sys.stderr.flush()




from raptor.build import add_buildpack, App
from raptor.utils import tmpdir


def test_buildpack_detected():
    with tmpdir():
        add_buildpack('https://github.com/heroku/heroku-buildpack-nodejs.git')
        add_buildpack('https://github.com/heroku/heroku-buildpack-python.git')
        app = App('node_app', 'https://bitbucket.org/btubbs/vr_node_example',
                  vcs_type='hg')
        app.update('tip')

        assert app.buildpack.basename == 'heroku-buildpack-nodejs'



import os
import time

import suds.xsd.doctor
import suds.client
from suds.plugin import MessagePlugin
from suds import WebFault

from . import base


# Suds has broken array marshaling.  See these links:
# http://stackoverflow.com/questions/3519818/suds-incorrect-marshaling-of-array-of-arrays
# https://fedorahosted.org/suds/ticket/340
class FixArrayPlugin(MessagePlugin):
    def marshalled(self, context):
        command = context.envelope.getChild('Body').getChildren()[0]
        # TODO: instead of blacklisting the affected types here, check the
        # actual WSDL and fix up any *ArrayArray types.
        affected = ('addNodes',
                    'addDrainingNodes',
                    'removeNodes',
                    'removeDrainingNodes',
                    'disableNodes',
                    'enableNodes',
                    'addPool',
                   )
        if command.name in affected:
            context.envelope.addPrefix('xsd', 'http://www.w3.org/1999/XMLSchema')
            # SO UGLY :(
            if command.name in ('addPool', 'disableNodes'):
                values = command.getChild('nodes')
            else:
                values = command.getChild('values')
            values.set('SOAP-ENC:arrayType', 'xsd:list[1]')
            values.set('xsi:type', 'SOAP-ENC:Array')
            item = values[0]
            item.set('SOAP-ENC:arrayType', 'xsd:list[%s]' % len(item.children))
            item.set('xsi:type', 'SOAP-ENC:Array')


class StingrayBalancer(base.Balancer):
    def __init__(self, config):
        self.url = config['URL']
        imp = suds.xsd.doctor.Import('http://schemas.xmlsoap.org/soap/encoding/')
        imp.filter.add('http://soap.zeus.com/zxtm/1.0/')
        doctor = suds.xsd.doctor.ImportDoctor(imp)

        # zxtm_pool.wsdl must be present in the same directory as this file.
        here = os.path.dirname(os.path.realpath(__file__))
        wsdl = os.path.join(here, 'stingray_pool.wsdl')
        self.client = suds.client.Client(
            'file:' + wsdl,
            username=config['USER'], password=config['PASSWORD'],
            location=self.url, plugins=[doctor, FixArrayPlugin()])

        # All pool names will be prefixed with this string.
        self.pool_prefix = config.get('POOL_PREFIX', '')

        # Stingray has separate calls for disableNodes and removeNodes.  The
        # latter will interrupt current connections.  To minimize disruption,
        # we'll call disableNodes first, wait a configurable amount of time,
        # and then call removeNodes.
        self.grace_period = config.get('GRACE_PERIOD', 2)

    def _call_node_func(self, func, pool, nodes):
        # Generic function for calling any of the Stingray pool functions that
        # accept an array of pools, and an arrayarray of nodes.  This function
        # will take a single pool and nodelist and do all the necessary
        # wrapping.
        nodes_wrapper = self.client.factory.create('StringArrayArray')
        nodes_array = self.client.factory.create('StringArray')
        nodes_array.item = nodes
        nodes_wrapper.item = [nodes_array]
        func([self.pool_prefix + pool], nodes_wrapper)

    def add_nodes(self, pool, nodes):
        # Stingray will kindly avoid creating duplicates if you submit a node
        # that is already in the pool.
        try:
            self._call_node_func(self.client.service.addNodes, pool, nodes)
        except WebFault as wf:
            if 'Unknown pool' in wf.message:
                # If pool doesn't exist, create it.
                self._call_node_func(self.client.service.addPool, pool, nodes)
            else:
                raise

    def delete_nodes(self, pool, nodes):
        try:
            self._call_node_func(self.client.service.disableNodes, pool, nodes)
            # wait <grace_period> seconds for connections to finish before
            # zapping nodes completely.
            time.sleep(self.grace_period)
            self._call_node_func(self.client.service.removeNodes, pool, nodes)
        except WebFault as wf:
            if 'Unknown pool' in wf.message:
                # If you try to delete nodes from a pool, and it doesn't exist,
                # that's fine.
                pass
            else:
                raise

    def get_nodes(self, pool):
        try:
            # get just the first item from the arrayarray
            nodes = self.client.service.getNodes([self.pool_prefix + pool])[0]
        except WebFault as wf:
            if 'Unknown pool' in wf.message:
                return []
            else:
                raise
        # convert the sax text things into real strings
        return [str(n) for n in nodes]



from collections import defaultdict
import logging

from . import base

class DummyBalancer(base.Balancer):
    """
    Fake balancer class for local development.
    """
    def __init__(self, config):
        self.config = config
        self.pools = defaultdict(set)

    def add_nodes(self, pool, nodes):
        self.pools[pool].update(nodes)
        self.log_pool(pool)

    def delete_nodes(self, pool, nodes):
        self.pools[pool].difference_update(nodes)
        self.log_pool(pool)

    def get_nodes(self, pool):
        return self.pools[pool]

    def log_pool(self, pool):
        msg = 'Dummy Pool "%s": %s' % (pool, list(self.pools[pool]))
        logging.info(msg)



import datetime
import unittest
import json

import redis
import pytest

from raptor.models import Host, Proc, ProcError
from raptor.utils import parse_redis_url, utc
from raptor.tests import FakeRPC


def test_host_init_rpc():
    server = FakeRPC()
    host = Host('somewhere', server)
    assert host.supervisor is server.supervisor


def test_host_init_port():
    host = Host('somewhere', 9001)
    assert str(host.rpc) == '<ServerProxy for somewhere:9001/RPC2>'


def test_host_init_redis_url():
    server = FakeRPC()
    host = Host('somewhere', server, redis_or_url='redis://localhost:6379/0')
    assert isinstance(host.redis, redis.StrictRedis)


def test_get_procs_len():
    server = FakeRPC()
    host = Host('somewhere', server)
    procs = host.get_procs()
    assert len(procs) == 2

class FakeProcCase(unittest.TestCase):
    def setUp(self):
        self.server = FakeRPC()
        self.host = Host('somewhere', self.server)
        self.dummyproc = self.host.get_proc('dummyproc')
        self.nodeproc = self.host.get_proc('node_example-v2-local-f96054b7-web-5003')

    def test_hostname(self):
        assert self.dummyproc.host.name == 'somewhere'

    def test_name(self):
        assert self.dummyproc.name == 'dummyproc'

    def test_start(self):
        ts = datetime.datetime(2012, 12, 19, 6, 19, 46, tzinfo=utc)
        assert self.dummyproc.start_time == ts

    def test_stop(self):
        ts = datetime.datetime(2012, 12, 19, 6, 19, 46, tzinfo=utc)
        assert self.dummyproc.stop_time == ts

    def test_now(self):
        ts = datetime.datetime(2012, 12, 19, 6, 19, 46, tzinfo=utc)
        assert self.dummyproc.now == ts

    def test_description(self):
        assert self.dummyproc.description == 'pid 5556, uptime 16:05:53'

    def test_group(self):
        assert self.dummyproc.group == 'dummyproc'

    def test_logfile(self):
        assert self.dummyproc.logfile == '/var/log/supervisor/dummyproc-stdout---supervisor-cYv5Q2.log'

    def test_pid(self):
        assert self.dummyproc.pid == 5556

    def test_spawnerr(self):
        assert self.dummyproc.spawnerr == ''

    def test_statename(self):
        assert self.dummyproc.statename == 'RUNNING'

    def test_stderr_logfile(self):
        assert self.dummyproc.stderr_logfile == '/tmp/pub.log'

    def test_stdout_logfile(self):
        assert self.dummyproc.stdout_logfile == '/var/log/supervisor/dummyproc-stdout---supervisor-cYv5Q2.log'

    def test_as_dict(self):
        assert self.dummyproc.as_dict() == {
             'app_name': 'dummyproc',
             'description': 'pid 5556, uptime 16:05:53',
             'exitstatus': 0,
             'group': 'dummyproc',
             'hash': 'UNKNOWN',
             'host': 'somewhere',
             'id': 'somewhere-dummyproc',
             'jsname': 'dummyproc',
             'logfile': '/var/log/supervisor/dummyproc-stdout---supervisor-cYv5Q2.log',
             'name': 'dummyproc',
             'now': '2012-12-19T06:19:46+00:00',
             'pid': 5556,
             'port': 0,
             'proc_name': 'dummyproc',
             'recipe_name': 'UNKNOWN',
             'spawnerr': '',
             'start_time': '2012-12-19T06:19:46+00:00',
             'state': 20,
             'statename': 'RUNNING',
             'stderr_logfile': '/tmp/pub.log',
             'stdout_logfile': '/var/log/supervisor/dummyproc-stdout---supervisor-cYv5Q2.log',
             'stop_time': '2012-12-19T06:19:46+00:00',
             'version': 'UNKNOWN'}

    def test_app_name(self):
        assert self.nodeproc.app_name == 'node_example'

    def test_version(self):
        assert self.nodeproc.version == 'v2'

    def test_recipe_name(self):
        assert self.nodeproc.recipe_name == 'local'

    def test_hash(self):
        assert self.nodeproc.hash == 'f96054b7'

    def test_proc_name(self):
        assert self.nodeproc.proc_name == 'web'

    def test_port(self):
        assert self.nodeproc.port == 5003

    def test_as_node(self):
        assert self.nodeproc.as_node() == 'somewhere:5003'

    def test_shortname(self):
        assert self.nodeproc.shortname() == 'node_example-v2-web'


def test_datetime_none():
    server = FakeRPC()
    server.supervisor.process_info['dummyproc']['now'] = 0
    host = Host('somewhere', server)
    proc = host.get_proc('dummyproc')
    assert proc.now is None


class RedisCacheTests(unittest.TestCase):
    def setUp(self):
        self.server = FakeRPC()
        self.supervisor = self.server.supervisor
        self.redis = redis.StrictRedis(**parse_redis_url('redis://localhost:6379/15'))
        self.host = Host('somewhere', self.server,
                         redis_or_url=self.redis)

    def tearDown(self):
        # self.redis.delete(self.host.cache_key)
        pass

    def test_proc_get_cache_set(self):
        # Ensure that single-proc data fetched from host is saved to cache
        self.host.get_proc('dummyproc', check_cache=True)
        cached = self.redis.hget(self.host.cache_key, 'dummyproc')
        assert json.loads(cached) == self.supervisor.process_info['dummyproc']

    def test_procs_get_cache_set(self):
        # Ensure that full data fetched from host is saved to cache
        self.host.get_procs(check_cache=True)
        cached = self.redis.hgetall(self.host.cache_key)
        parsed = {k: json.loads(v) for k, v in cached.items()}
        assert parsed == self.supervisor.process_info

    def test_get_procs_uses_cache(self):
        self.host.get_procs(check_cache=True)
        # If we actually hit the RPC, an exception will be raised
        self.supervisor.exception = AssertionError('cache not used')
        self.host.get_procs(check_cache=True)

    def test_get_proc_uses_cache(self):
        self.host.get_proc('dummyproc', check_cache=True)
        # If we actually hit the RPC, an exception will be raised
        self.supervisor.exception = AssertionError('cache not used')
        self.host.get_proc('dummyproc', check_cache=True)

    def test_full_set_partial_get(self):
        # Ensure that data cached from a full get is used when just requesting
        # a single proc
        procs = self.host.get_procs(check_cache=True)
        # If we actually hit the RPC, an exception will be raised
        self.supervisor.exception = AssertionError('cache not used')
        self.host.get_proc('dummyproc', check_cache=True)

    def test_absent_proc_decached(self):
        # Test that if a proc is in the cache but not returned from the host
        # when get_procs is called, it gets removed from the cache
        data = dict(self.supervisor.process_info['dummyproc'])
        data['group'] = data['name'] = 'deadproc'
        self.redis.hset(self.host.cache_key, 'deadproc', json.dumps(data))

        # Requesting just this proc should return data
        assert self.host.get_proc('deadproc', check_cache=True).name == 'deadproc'

        # But requesting all procs, with deadproc absent from the Supervisor
        # data, should clear him from cache
        all_procs = [p.name for p in self.host.get_procs(check_cache=False)]
        assert 'deadproc' not in all_procs

    def test_nonexistent_proc_raises_proc_error(self):
        with pytest.raises(ProcError):
            self.host.get_proc('nonexistent')



import json

from mock import Mock, call

from raptor import publisher
from raptor.tests import FakeRPC
from raptor.models import Host, Proc

def get_tick():
    """
    Create and return a faked "tick" event as would be emitted from Supervisor.
    """
    epayload = 'when:1356028501'
    eheaders = {
        'ver': '3.0',
        'server': 'supervisor',
        'serial': '21',
        'pool': 'listener',
        'poolserial': '10',
        'eventname': 'TICK_60',
        'len': str(len(epayload)),
    }
    return publisher.Event(eheaders, epayload, 'somewhere')

# tick events just update cache
def test_ticks_get_fresh_data():
    server = Mock()
    host = Host('somewhere', server)
    tick = get_tick()
    try:
        publisher.handle_event(tick, host, 'fake_channel')
    except TypeError:
        # TypeError: 'Mock' object is not iterable
        # This happens after the bit we're testing, so ignore it.
        pass
    assert server.supervisor.mock_calls == [call.getAllProcessInfo()]


class MockContextManager(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.mock = Mock()
    def __enter__(self):
        return self.mock
    def __exit__(self, *_):
        pass

def test_events_cache_all_procs():
    server = FakeRPC()
    host = Host('somewhere', server)
    host.redis = Mock()
    host.redis.pipeline.return_value = pipe = MockContextManager()
    host.redis.hkeys.return_value = []
    tick = get_tick()
    publisher.handle_event(tick, host, 'fake_channel')
    calls = pipe.mock.mock_calls
    assert calls[0][0] == 'delete'
    cache_key, hashdict = calls[1][1]
    assert cache_key == 'host_procs:somewhere'
    for k in hashdict:
        assert k in ('dummyproc', 'node_example-v2-local-f96054b7-web-5003')
        # Make sure the saved values are json dicts
        procdata = json.loads(hashdict[k])
        assert isinstance(procdata, dict)
    assert calls[2] == call.expire('host_procs:somewhere', 600)
    assert calls[3] == call.execute()


def test_procstate_events_published():
    server = FakeRPC()
    host = Host('somewhere', server)
    host.redis = Mock()
    host.redis.pipeline.return_value = MockContextManager()
    host.redis.hkeys.return_value = []
    epayload = 'processname:dummyproc groupname:dummyproc from_state:STOPPED tries:0'
    eheaders = {
        'ver': '3.0',
        'server': 'supervisor',
        'serial': '21',
        'pool': 'listener',
        'poolserial': '10',
        'eventname': 'PROCESS_STATE_RUNNING',
        'len': str(len(epayload)),
    }
    ev = publisher.Event(eheaders, epayload, host.name)
    publisher.handle_event(ev, host, 'fake_channel')

    _, args, kwargs = host.redis.publish.mock_calls[0]
    assert args[0] == 'fake_channel'

    expected_data = Proc(host,
                         dict(host.supervisor.process_info['dummyproc'])).as_dict()
    expected_data['event'] = 'PROCESS_STATE_RUNNING'
    published_data = json.loads(args[1])
    assert expected_data == published_data


def test_removal_events_published():
    server = FakeRPC()
    host = Host('somewhere', server)
    host.redis = Mock()
    host.redis.pipeline.return_value = MockContextManager()
    host.redis.hkeys.return_value = []
    epayload = 'groupname:dummyproc'
    eheaders = {
        'ver': '3.0',
        'server': 'supervisor',
        'serial': '21',
        'pool': 'listener',
        'poolserial': '10',
        'eventname': 'PROCESS_GROUP_REMOVED',
        'len': str(len(epayload)),
    }
    ev = publisher.Event(eheaders, epayload, host.name)
    publisher.handle_event(ev, host, 'fake_channel')

    _, args, kwargs = host.redis.publish.mock_calls[0]
    assert args[0] == 'fake_channel'

    parsed = json.loads(args[1])
    assert parsed['name'] == 'dummyproc'
    assert parsed['event'] == 'PROCESS_GROUP_REMOVED'
    assert 'time' in parsed


def test_removals_include_id():
    server = FakeRPC()
    host = Host('somewhere', server)
    host.redis = Mock()
    host.redis.pipeline.return_value = MockContextManager()
    host.redis.hkeys.return_value = []
    epayload = 'groupname:dummyproc'
    eheaders = {
        'ver': '3.0',
        'server': 'supervisor',
        'serial': '21',
        'pool': 'listener',
        'poolserial': '10',
        'eventname': 'PROCESS_GROUP_REMOVED',
        'len': str(len(epayload)),
    }
    ev = publisher.Event(eheaders, epayload, host.name)
    publisher.handle_event(ev, host, 'fake_channel')

    _, args, kwargs = host.redis.publish.mock_calls[0]
    parsed = json.loads(args[1])
    assert parsed['id'] == 'somewhere-dummyproc'

def test_removals_include_parsed_procname():

    server = FakeRPC()
    host = Host('somewhere', server)
    host.redis = Mock()
    host.redis.pipeline.return_value = MockContextManager()
    host.redis.hkeys.return_value = []
    epayload = 'groupname:node_example-v2-local-f96054b7-web-5003'
    eheaders = {
        'ver': '3.0',
        'server': 'supervisor',
        'serial': '21',
        'pool': 'listener',
        'poolserial': '10',
        'eventname': 'PROCESS_GROUP_REMOVED',
        'len': str(len(epayload)),
    }
    ev = publisher.Event(eheaders, epayload, host.name)
    publisher.handle_event(ev, host, 'fake_channel')

    _, args, kwargs = host.redis.publish.mock_calls[0]
    parsed = json.loads(args[1])
    assert parsed['app_name'] == 'node_example'
    assert parsed['hash'] == 'f96054b7'
    assert parsed['proc_name'] == 'web'
    assert parsed['recipe_name'] == 'local'
    assert parsed['version'] == 'v2'
    assert parsed['port'] == 5003



#!/usr/bin/python
from setuptools import setup, find_packages

setup(
    name='raptor',
    version='2.15',
    author='Brent Tubbs',
    author_email='brent.tubbs@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Fabric==1.4.0',
        'paramiko>=1.8.0,<2.0',
        'suds==0.4',
        'supervisor==3.0b2-dev-vr4',
        'redis>=2.6.2',
        'isodate>=0.4.4',
    ],
    entry_points={
        'console_scripts': [
            'proc_publisher = raptor.publisher:main',
        ]
    },
    # Custom version of supervisor that supports proc add/delete events, and
    # logging to both syslog and local files.
    dependency_links = ['https://bitbucket.org/yougov/velociraptor/downloads/supervisor-3.0b2-dev-vr4.tar.gz#egg=supervisor-3.0b2-dev-vr4'],
    description=('Libraries and command line tools for deploying with '
                 'Velociraptor'),
)



import errno
import posixpath
import re

from . import base

# Template for 'upstream' directive in nginx config.
UPSTREAM_TPL = """
upstream %(name)s {
%(lines)s
}
"""


def pool_to_str(name, nodes):
    """
    Given a list of nodes, return a string containing a properly-formatted
    'upstream' section for nginx config.
    """
    return UPSTREAM_TPL % {
        'name': name,
        'lines': '\n'.join(['  server %s;' % n for n in nodes])
    }


def str_to_pool(upstream):
    """
    Given a string containing an nginx upstream section, return the pool name
    and list of nodes.
    """
    name = re.search('upstream +(.*?) +{', upstream).group(1)
    nodes = re.findall('server +(.*?);', upstream)
    return name, nodes


class NginxBalancer(base.SshBasedBalancer):
    """
    A Velociraptor balancer backend for writing nginx config on remote servers.
    Uses sftp and ssh for writing files and running reload cmds.
    """

    def __init__(self, config):
        self.include_dir = config.get('include_dir',
                                      '/etc/nginx/sites-enabled/')
        self.reload_cmd = config.get('reload_cmd', '/etc/init.d/nginx reload')
        super(NginxBalancer, self).__init__(config)

    def _get_host_nodes(self, host, pool):
        # Return set of nodes currently configured in a given host and pool

        path = posixpath.join(self.include_dir, pool + '.conf')
        try:
            contents = self._read_file(host, path)
            poolname, nodes = str_to_pool(contents)
        except IOError as e:
            # It's OK if the file doesn't exist.  But other IOErrors should be
            # raised normally.
            if e.errno == errno.ENOENT:
                nodes = []
            else:
                raise
        return set(nodes)

    def _set_host_nodes(self, host, pool, nodes):
        path = posixpath.join(self.include_dir, pool + '.conf')
        contents = pool_to_str(pool, nodes)
        self._write_file(host, path, contents)
